/* cache stuff for mod_auth_any */
#include "maa_cache.h"

typedef struct account_info {
    char *user;
    char *pass;
    char *name;
    time_t expire_when;
} auth_account_rec;

static auth_account_rec accounts[MAA_MAX_CACHE_ENTRIES];
static int numcacheentries=0;


/* perform timeouts */
void do_cache_timeouts(const time_t now,int fullcache,request_rec *r) {
    int i,delme=MAA_MAX_CACHE_ENTRIES;
    auth_account_rec n;
    time_t oldest=0;
    time_t expiry;
    if(MAA_DEBUGLVL>2) 
        ap_log_rerror(MAA_DEBUG,"About to init");
    
    if(numcacheentries==0) {
        for(i=0;i<MAA_MAX_CACHE_ENTRIES;i++) {
            accounts[i].user=NULL;
            accounts[i].expire_when=0;
        }
        return; /* we might as well; with an empty cache we still have much work to do */
    }
    if(MAA_DEBUGLVL>2) 
        ap_log_rerror(MAA_DEBUG,"After cache init");

    for(i=0;i<MAA_MAX_CACHE_ENTRIES;i++) {
        n=accounts[i];
        if(n.user==NULL) continue;
        if(MAA_DEBUGLVL>3) 
            ap_log_rerror(MAA_DEBUG,"Doing timeout for entry %d of %d entries",i,numcacheentries);
        expiry=n.expire_when;
        if(fullcache==1 && (oldest==0 || expiry<oldest)) {
            oldest=expiry;
            delme=i;
        }
        if(MAA_DEBUGLVL>3) 
            ap_log_rerror(MAA_DEBUG,"Testing time-out entry %d of %d entries",i,numcacheentries);
        if(expiry<now) {
            if(MAA_DEBUGLVL>3) 
                ap_log_rerror(MAA_DEBUG,"Timing out %s named %s with timeout=%ld",n.user,n.name,n.expire_when);
            free(n.name);
            free(n.user);
            free(n.pass);
            accounts[i].user=NULL;
            numcacheentries--;
            fullcache=0;
        }
        if(MAA_DEBUGLVL>3) 
            ap_log_rerror(MAA_DEBUG,"Done testing timeout for entry %d of %d entries",i,numcacheentries);
    }
    if(MAA_DEBUGLVL>2) 
        ap_log_rerror(MAA_DEBUG,"After timeout checks. delme=%d max cache=%d",delme,MAA_MAX_CACHE_ENTRIES);

    /* delme shouldn't be equal to max cache entries if we found an "oldest"
     * record 
     */
    if(fullcache==1 && delme<MAA_MAX_CACHE_ENTRIES) {
        if(MAA_DEBUGLVL>3) 
            ap_log_rerror(MAA_DEBUG,"Forcing timeout for entry %d of %d entries",delme,numcacheentries);
        n=accounts[delme];
        free(n.user);
        free(n.pass);
        free(n.name);
        accounts[delme].user=NULL;
        numcacheentries--;
    }
    if(MAA_DEBUGLVL>2) 
        ap_log_rerror(MAA_DEBUG,"After forced time-out. delme=%d max cache=%d",delme,MAA_MAX_CACHE_ENTRIES);
    
    return;
} /* do_cache_timeouts */


/* lookup user and pass in cache */
char * lookup_user(
        request_rec *r,
        char *user,
        const char *pass
        ) {
    char *found;
    char *u;
    char *p;
    size_t maxu, lu;
    int i;
    time_t now;
    size_t maxp, lp;
    lu=strnlen(user,MAA_BUFLEN);

    found=NULL;
    (void)time(&now);
    if(MAA_DEBUGLVL>6) 
        ap_log_rerror(MAA_DEBUG,"Time before timeout: %ld",now);
    do_cache_timeouts(now,0,r);
    (void)time(&now);
    if(MAA_DEBUGLVL>6) 
        ap_log_rerror(MAA_DEBUG,"Time after timeout: %ld",now);

#ifdef MAA_USE_MD5
maxp=MAA_PASSSIZE;
lp=MAA_PASSSIZE;
#else
    lp=strnlen(pass,MAA_BUFLEN);
#endif
    for(i=0;i<MAA_MAX_CACHE_ENTRIES;i++) {
        if(accounts[i].user==NULL) continue;
        u=accounts[i].user;
            ap_log_rerror(MAA_DEBUG,"%s in cache",u);
        p=accounts[i].pass;
        maxu=strnlen(u,MAA_BUFLEN);
#ifndef MAA_USE_MD5
        maxp=strnlen(p,MAA_BUFLEN);
#endif
        if(MAA_DEBUGLVL>3 && lp!=maxp) 
            ap_log_rerror(MAA_DEBUG,"%s in cache lu=%d lp=%d maxu=%d maxp=%d %s %s",u,lu,lp,maxu,maxp,p,pass);
        if(MAA_DEBUGLVL>3) 
            ap_log_rerror(MAA_DEBUG,"%s in cache lu=%d lp=%d maxu=%d maxp=%d",u,lu,lp,maxu,maxp);
        if(maxu!=lu || maxp!=lp || memcmp(pass,p,maxp)!=0 || strncmp(user,u,maxu)!=0) {
            /* mismatch */
            continue;
        }
        /* match */
        if(MAA_DEBUGLVL>1) 
            ap_log_rerror(MAA_DEBUG,"%s found in cache, returning %d",user,numcacheentries);
        
        found=accounts[i].name;
        break;
    }
    if(MAA_DEBUGLVL>0 && found==NULL)
        ap_log_rerror(MAA_DEBUG,"%s not found in cache: %d",user,numcacheentries);

    return found;
} /* lookup_user */


/* add user and pass to cache */
void update_cache(
        request_rec *r,
        char *user,
        const char *pass,
        char *name
        ) {
    auth_account_rec newnode;
    time_t now;
    int i;
    
    (void)time(&now);
    if(MAA_DEBUGLVL>2)
        ap_log_rerror(MAA_DEBUG,"Time before allocate: %ld",now);

    if(user==NULL) return;

    /*
    newnode=(auth_account_rec *)malloc(sizeof(auth_account_rec));
    if(newnode==NULL) {
        ap_log_rerror(APLOGERROR,"Unable to allocate memory for new node in cache");
        return;
    }
    */
    newnode.user=NULL;
    newnode.pass=NULL;
    newnode.name=NULL;
    newnode.user=strndup(user,MAA_BUFLEN-1);
    if(user !=NULL && newnode.user==NULL) {
        ap_log_rerror(MAA_ERROR,"Unable to allocate memory for new node data in cache");
        return;
    }
#ifdef MAA_USE_MD5
    newnode.pass=(char *)malloc(MAA_PASSSIZE);
    if(pass !=NULL && newnode.pass==NULL) {
        ap_log_rerror(MAA_ERROR,"Unable to allocate memory for new node data in cache");
        return;
    }
    memcpy(newnode.pass,pass,MAA_PASSSIZE);
#else
    newnode.pass=strndup(pass,MAA_BUFLEN-1);
#endif
    if(pass !=NULL && newnode.pass==NULL) {
        ap_log_rerror(MAA_ERROR,"Unable to allocate memory for new node data in cache");
        return;
    }
    newnode.name=strndup(name,MAA_BUFLEN-1);
    if(name !=NULL && newnode.name==NULL) {
        ap_log_rerror(MAA_ERROR,"Unable to allocate memory for new node data in cache");
        return;
    }
    newnode.expire_when=now+MAA_CACHE_EXPIRE;

    if(numcacheentries>=MAA_MAX_CACHE_ENTRIES) {
    if(MAA_DEBUGLVL>0) 
        ap_log_rerror(MAA_DEBUG,"mod_auth_any: cache: full, expiring %d of %d max",numcacheentries,MAA_MAX_CACHE_ENTRIES);
        do_cache_timeouts(now,1,r);
    }
    for(i=0;i<MAA_MAX_CACHE_ENTRIES;i++) {
        if(accounts[i].user==NULL) {
		if(MAA_DEBUGLVL>0) 
			ap_log_rerror(MAA_DEBUG,"mod_auth_any: cache: inserting at %d",i);
		accounts[i].user=newnode.user;
		accounts[i].pass=newnode.pass;
		accounts[i].name=newnode.name;
		accounts[i].expire_when=newnode.expire_when;
		if(MAA_DEBUGLVL>2) 
			ap_log_rerror(MAA_DEBUG,"mod_auth_any: cache: inserted %s %s",accounts[i].name,accounts[i].user);
		numcacheentries++;
            break;
        }
    }

    if(MAA_DEBUGLVL>2) {
        (void)time(&now);
        ap_log_rerror(MAA_DEBUG,"Time after allocate: %ld",now);
    }

    return;

} /* update_cache */



/* eof */
